#include <stdio.h>
#include "kernels.h"

// generate a matrix for a 2d Poisson problem
// discretized on an NxN grid with a 5-point stencil
void initialize_example(int N,
                        int * num_rows,
                        int * num_cols,
                        int * num_entries,
                        int   ** row_offsets,
                        int   ** column_indices,
                        float ** values,
                        float ** x,
                        float ** y)
{
  *num_rows = N*N;
  *num_cols = N*N;
  *num_entries = 5*N*N - 4*N;

  // allocate row_pointer array 
  *row_offsets = (int *) malloc( (*num_rows + 1) * sizeof(int) );
  
  // allocate the column indices and values array
  *column_indices = (int *)   malloc( *num_entries * sizeof(int)   );
  *values         = (float *) malloc( *num_entries * sizeof(float) );
                                               
  (*row_offsets)[0] = 0;                                               

  int nz = 0;

  for(int i = 0; i < N; i++)
  {
    for(int j = 0; j < N; j++)
    {
      int indx = N*i + j;

      if (i > 0)
      {
        (*column_indices)[nz] = indx - N; (*values)[nz] = -1; nz++;
      }

      if (j > 0)
      {
        (*column_indices)[nz] = indx - 1; (*values)[nz] = -1; nz++;
      }

      (*column_indices)[nz] = indx; (*values)[nz] = 4; nz++;

      if (j < N - 1)
      {
        (*column_indices)[nz] = indx + 1; (*values)[nz] = -1; nz++;
      }

      if (i < N - 1)
      {
        (*column_indices)[nz] = indx + N; (*values)[nz] = -1; nz++;
      }

      (*row_offsets)[indx + 1] = nz;
    }
  }
  
  // allocate x and y arrays
  *x = (float *) malloc( *num_cols * sizeof(float) );
  *y = (float *) malloc( *num_rows * sizeof(float) );

  for (int i = 0; i < *num_cols; i++)
    (*x)[i] = (31 * i) % 10;

  for (int i = 0; i < *num_rows; i++)
    (*y)[i] = 0;
}


int main(void)
{
  // initialize csr matrix and vectors on host
  int num_rows;
  int num_cols;
  int num_entries;
  int   * host_row_offsets;
  int   * host_column_indices;
  float * host_values;
  float * host_x;
  float * host_y;

  initialize_example(1024,
                     &num_rows, &num_cols, &num_entries,
                     &host_row_offsets, &host_column_indices, &host_values,
                     &host_x, &host_y);

  // copy csr matrix from host to device
  int   * device_row_offsets;
  int   * device_column_indices;
  float * device_values;
  float * device_x;
  float * device_y;
  cudaMalloc(&device_row_offsets,    (num_rows + 1) * sizeof(int));
  cudaMalloc(&device_column_indices, num_entries    * sizeof(int));
  cudaMalloc(&device_values,         num_entries    * sizeof(float));
  cudaMalloc(&device_x,              num_cols       * sizeof(float));
  cudaMalloc(&device_y,              num_rows       * sizeof(float));
  cudaMemcpy(device_row_offsets,    host_row_offsets,    (num_rows + 1) * sizeof(int),   cudaMemcpyHostToDevice);
  cudaMemcpy(device_column_indices, host_column_indices, num_entries    * sizeof(int),   cudaMemcpyHostToDevice);
  cudaMemcpy(device_values,         host_values,         num_entries    * sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(device_x,              host_x,              num_cols       * sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(device_y,              host_y,              num_rows       * sizeof(float), cudaMemcpyHostToDevice);

  // compute spmv on host
  csr_spmv_host(num_rows, num_cols, num_entries,
                host_row_offsets, host_column_indices, host_values,
                host_x, host_y);
  
  // compute spmv on device
  csr_spmv_device(num_rows, num_cols, num_entries,
                  device_row_offsets, device_column_indices, device_values,
                  device_x, device_y);

  // copy device_y to host
  float * device_y_on_host = (float *) malloc(num_rows * sizeof(float));
  cudaMemcpy(device_y_on_host, device_y, num_rows * sizeof(float), cudaMemcpyDeviceToHost);

  // compare host and device results
  int mismatches = 0;
  printf("Comparing CSR SpMV results...\n");
  for (int row = 0; row < num_rows; row++)
  {
    if (host_y[row] != device_y_on_host[row])
    {
      if (mismatches == 0)
        printf("%12s %12s %12s\n", "Row", "CPU result", "GPU result");

      if (mismatches < 20)
        printf("%12d %12.4f %12.4f\n", row, host_y[row], device_y_on_host[row]);

      mismatches++;
    }
  }
  
  if (mismatches == 0)
  {
    printf("PASSED: host and device results are equal\n");
  }
  else
  {
    printf("FAILED: host and device results are not equal (%d mismatches)\n", mismatches);
    return 1;
  }

  // benchmark spmv on device
  printf("Benchmarking Device SpMV...\n");

  cudaEvent_t start;
  cudaEvent_t end;
  cudaEventCreate(&start); 
  cudaEventCreate(&end);
  cudaEventRecord(start,0);
  
  int iterations = 10;

  for (int i = 0; i < iterations; i++)
  {
    // compute spmv on device
    csr_spmv_device(num_rows, num_cols, num_entries,
                    device_row_offsets, device_column_indices, device_values,
                    device_x, device_y);
  }

  // compute time
  float milliseconds;
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);
  cudaEventElapsedTime(&milliseconds, start, end);
  
  milliseconds /= iterations;

  float gflops = 2 * num_entries / (1e6 * milliseconds);
  
  printf("Computed SpMV in %f milliseconds (%f GFLOP/s)\n", milliseconds, gflops);

  // free memory
  free(host_row_offsets);
  free(host_column_indices);
  free(host_values);
  free(host_x);        
  free(host_y);
  cudaFree(device_row_offsets);
  cudaFree(device_column_indices);
  cudaFree(device_values);
  cudaFree(device_x);        
  cudaFree(device_y);
  free(device_y_on_host);
  cudaEventDestroy(start);
  cudaEventDestroy(end);
}

