// CSR SpMV on device
__global__
void csr_kernel(const int num_rows,
                const int   * row_offsets,
                const int   * column_indices,
                const float * values,
                const float * x,
                      float * y)
{
  int row = blockDim.x * blockIdx.x + threadIdx.x;

  if (row < num_rows)
  {
    float sum = 0;

    int row_begin = row_offsets[row];
    int row_end   = row_offsets[row + 1];

    for (int offset = row_begin; offset < row_end; offset++)
    {
      sum += values[offset] * x[column_indices[offset]];
    }

    y[row] = sum;
  }
}
  

// CSR SpMV on device
void csr_spmv_device(const int num_rows,
                     const int num_cols,
                     const int num_entries,
                     const int   * row_offsets,
                     const int   * column_indices,
                     const float * values,
                     const float * x,
                           float * y)
{
  // TODO launch CSR SpMV kernel 
  int block_size = 256;
  int grid_size  = (num_rows + (block_size - 1)) / block_size;

  csr_kernel<<<grid_size, block_size>>>(num_rows, row_offsets, column_indices, values, x, y);
}

// CSR SpMV on host
void csr_spmv_host(const int num_rows,
                   const int num_cols,
                   const int num_entries,
                   const int   * row_offsets,
                   const int   * column_indices,
                   const float * values,
                   const float * x,
                         float * y)
{
  for (int row = 0; row < num_rows; row++)
  {
    float sum = 0;

    int row_begin = row_offsets[row];
    int row_end   = row_offsets[row + 1];

    for (int offset = row_begin; offset < row_end; offset++)
    {
      sum += values[offset] * x[column_indices[offset]];
    }

    y[row] = sum;
  }
}

__global__
void ell_kernel(const int num_rows,
                const int num_entries_per_row,
                const int   * column_indices,
                const float * values,
                const float * x,
                      float * y)
{
  int row = blockDim.x * blockIdx.x + threadIdx.x;

  if (row < num_rows)
  {
    float sum = 0;

    for (int k = 0; k < num_entries_per_row; k++)
    {
      int offset = num_rows * k + row;
      int col    = column_indices[offset];

      if (col != -1)
        sum += values[offset] * x[col];
    }

    y[row] = sum;
  }
}


// ELL SpMV on device
void ell_spmv_device(const int num_rows,
                     const int num_cols,
                     const int num_entries,
                     const int num_entries_per_row,
                     const int   * column_indices,
                     const float * values,
                     const float * x,
                           float * y)
{
  // TODO launch ELL SpMV kernel 
  int block_size = 256;
  int grid_size  = (num_rows + (block_size - 1)) / block_size;

  ell_kernel<<<grid_size, block_size>>>(num_rows, num_entries_per_row, column_indices, values, x, y);
}

// ELL SpMV on host
void ell_spmv_host(const int num_rows,
                   const int num_cols,
                   const int num_entries,
                   const int num_entries_per_row,
                   const int   * column_indices,
                   const float * values,
                   const float * x,
                         float * y)
{
  for (int row = 0; row < num_rows; row++)
  {
    float sum = 0;

    for (int k = 0; k < num_entries_per_row; k++)
    {
      int offset = num_rows * k + row;
      int col    = column_indices[offset];

      if (col != -1)
        sum += values[offset] * x[col];
    }

    y[row] = sum;
  }
}

